@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-# In[3]:
+# In[2]:
 
 
 try:
@@ -12,10 +12,16 @@ except:
     pass
 
 
-# In[4]:
+# In[ ]:
 
 
-version = '0.108'
+
+
+
+# In[3]:
+
+
+version = '0.109'
 
 import re
 import tempfile
@@ -33,7 +39,7 @@ import time
 
 # # Helper Functions
 
-# In[5]:
+# In[4]:
 
 
 def is_testing():
@@ -43,7 +49,7 @@ def is_testing():
         return False
 
 
-# In[6]:
+# In[5]:
 
 
 def isnotebook():
@@ -54,7 +60,7 @@ def isnotebook():
     return False
 
 
-# In[7]:
+# In[6]:
 
 
 def testPrint(*args):
@@ -64,7 +70,7 @@ def testPrint(*args):
 
 # # Inspect
 
-# In[8]:
+# In[7]:
 
 
 import inspect
@@ -76,7 +82,7 @@ def debugFunctionName():
     return inspect.stack()[1][3]
 
 
-# In[9]:
+# In[8]:
 
 
 debugGetFileName(), debugGetLineNumber(), debugFunctionName()
@@ -84,7 +90,7 @@ debugGetFileName(), debugGetLineNumber(), debugFunctionName()
 
 # # Shell
 
-# In[10]:
+# In[9]:
 
 
 def shell(cmd):
@@ -97,18 +103,28 @@ def shell(cmd):
     return out
 
 
-# In[11]:
+# In[10]:
 
 
 shell_template = """/bin/bash << EOF 2>&1
 {cmd}
 EOF"""
 
-def bash(cmd, lineList=True):
-    result = subprocess.run(shell_template.format(cmd=cmd), shell=True, stdout=subprocess.PIPE)
-    out = result.stdout.decode()
+def bash(cmd, lineList=True, do_print=False, remove_last_empty_line=True):
+    from subprocess import Popen, PIPE, STDOUT
+    p = Popen(cmd, stdout=PIPE, stderr=STDOUT, shell=True)
+    out = ""
+    while True:
+        line = p.stdout.readline().decode()
+        out += line
 
-    if len(out) > 0 and out[-1] == "\n":
+        if do_print:
+            print(line, end="")
+
+        if not line:
+            break
+
+    if remove_last_empty_line and len(out) > 0 and out[-1] == "\n":
         out = out[:-1]
 
     if lineList:
@@ -116,16 +132,24 @@ def bash(cmd, lineList=True):
     else:
         return out
 
+    return result
+
+
+# In[ ]:
+
+
+
+
 
 # # Timing
 
-# In[12]:
+# In[11]:
 
 
 timeSinceLastCallDict = {}
 
 
-# In[13]:
+# In[12]:
 
 
 def timeSinceLastCall(name):
@@ -138,14 +162,14 @@ def timeSinceLastCall(name):
     return name + ": " + time_s
 
 
-# In[14]:
+# In[13]:
 
 
 if is_testing():
     print(timeSinceLastCall('test marker'))
 
 
-# In[15]:
+# In[14]:
 
 
 import time
@@ -173,7 +197,7 @@ if is_testing():
 
 # # Requiremtns
 
-# In[16]:
+# In[15]:
 
 
 requirements = []
@@ -205,7 +229,7 @@ import yaml
 
 # # Strings
 
-# In[17]:
+# In[16]:
 
 
 def sFloat0(f): return "{:0.0f}".format(f)
@@ -235,7 +259,7 @@ if is_testing():
     print(sFloatSci5(1.234567))
 
 
-# In[18]:
+# In[17]:
 
 
 def strUuid4():
@@ -246,14 +270,14 @@ def strUuid4():
     return uuid
 
 
-# In[19]:
+# In[18]:
 
 
 if is_testing():
     print(strUuid4())
 
 
-# In[20]:
+# In[19]:
 
 
 def strRand(n, num=True, upperCase=True, lowerCase=True):
@@ -265,7 +289,7 @@ def strRand(n, num=True, upperCase=True, lowerCase=True):
     return ''.join(random.choice(pool) for _ in range(n))
 
 
-# In[21]:
+# In[20]:
 
 
 if is_testing():
@@ -274,7 +298,7 @@ if is_testing():
     print(strRand(20, upperCase=False, lowerCase=False))
 
 
-# In[22]:
+# In[21]:
 
 
 def strMd5Pickle4(obj):
@@ -284,14 +308,14 @@ def strMd5Pickle4(obj):
     return md5.hexdigest()
 
 
-# In[23]:
+# In[22]:
 
 
 if is_testing():
     print(strMd5Pickle4(['a', 'b']))
 
 
-# In[24]:
+# In[23]:
 
 
 def strMd5String(s):
@@ -300,7 +324,7 @@ def strMd5String(s):
     return md5.hexdigest()
 
 
-# In[25]:
+# In[24]:
 
 
 if is_testing():
@@ -309,7 +333,7 @@ if is_testing():
 
 # # Multiprocess / Thread
 
-# In[26]:
+# In[25]:
 
 
 def mapThread(func, args, n=8):
@@ -334,14 +358,14 @@ if is_testing():
 
 # # Lists
 
-# In[27]:
+# In[26]:
 
 
 def liChunks(l, chunk_size):
     return [l[i:i + chunk_size] for i in range(0, len(l), chunk_size)]
 
 
-# In[28]:
+# In[27]:
 
 
 if is_testing():
@@ -350,7 +374,7 @@ if is_testing():
     print(liChunks(['a', 'b', 'c', 'd', 'e', 'f'], 3))
 
 
-# In[29]:
+# In[28]:
 
 
 def liSplit(a, number_of_splits):
@@ -358,7 +382,7 @@ def liSplit(a, number_of_splits):
     return [a[start:end] for start, end in zip(idxs[:-1], idxs[1:])]
 
 
-# In[30]:
+# In[29]:
 
 
 if is_testing():
@@ -370,7 +394,7 @@ if is_testing():
     print(liSplit(['a', 'b', 'c', 'd', 'e'], 6))
 
 
-# In[31]:
+# In[30]:
 
 
 def liJoin(l):
@@ -380,14 +404,14 @@ def liJoin(l):
     return out
 
 
-# In[32]:
+# In[31]:
 
 
 if is_testing():
     print(liJoin([['a', 'b', 'c'], ['d', 'e', 'f']]))
 
 
-# In[33]:
+# In[32]:
 
 
 def liFlatten(l):
@@ -404,7 +428,7 @@ def liFlatten(l):
         return l
 
 
-# In[34]:
+# In[33]:
 
 
 if is_testing():
@@ -414,7 +438,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[35]:
+# In[34]:
 
 
 regex_cache = {}
@@ -424,35 +448,35 @@ def sListFilter(l, re_include=".*", re_exclude='^\b$'):
     return [e for e in l if regex_cache[re_include].search(e) and not regex_cache[re_exclude].search(e)]
 
 
-# In[36]:
+# In[35]:
 
 
 if is_testing():
     sListFilter(['good x', 'other good x', 'without the letter before y'])
 
 
-# In[37]:
+# In[36]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[38]:
+# In[37]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[39]:
+# In[38]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[40]:
+# In[39]:
 
 
 def lCsv(l, force_oneline=False):
@@ -468,21 +492,21 @@ def lCsv(l, force_oneline=False):
     return ", ".join([str(e) for e in l])
 
 
-# In[41]:
+# In[40]:
 
 
 if is_testing():
     print(lCsv(['a', 'b', 1, 3.14]))
 
 
-# In[42]:
+# In[41]:
 
 
 if is_testing():
     print(lCsv([['a', 'b'], 1, 3.14]))
 
 
-# In[43]:
+# In[42]:
 
 
 if is_testing():
@@ -491,7 +515,7 @@ if is_testing():
 
 # # Text files
 
-# In[44]:
+# In[43]:
 
 
 def txtread(path):
@@ -512,7 +536,7 @@ def txtwrite(path, txt):
             return f.write(txt)
 
 
-# In[45]:
+# In[44]:
 
 
 if is_testing():
@@ -521,21 +545,21 @@ if is_testing():
 
 # # YAML files
 
-# In[46]:
+# In[45]:
 
 
 def yamlwrite(path, obj):
-    txtwrite(path, yaml.dump(obj))
+    txtwrite(path, yaml.dump(obj, sort_keys=False))
 
 
-# In[47]:
+# In[46]:
 
 
 def yamlread(path):
     return yaml.safe_load(txtread(path=path))
 
 
-# In[48]:
+# In[47]:
 
 
 if is_testing():
@@ -548,7 +572,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[49]:
+# In[48]:
 
 
 regex_cache = {}
@@ -560,35 +584,35 @@ def sListFilter(listOfStrings, re_include=".*", re_exclude='^\b$'):
 
 # ### Examples
 
-# In[50]:
+# In[49]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y']))
 
 
-# In[51]:
+# In[50]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[52]:
+# In[51]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[53]:
+# In[52]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[54]:
+# In[53]:
 
 
 if is_testing():
@@ -597,7 +621,7 @@ if is_testing():
 
 # # GCP Google Cloud Platform
 
-# In[55]:
+# In[54]:
 
 
 import warnings
@@ -621,7 +645,7 @@ def _gsBlob(path):
     return blob
 
 
-# In[56]:
+# In[55]:
 
 
 def gsGsToHttp(path):
@@ -631,7 +655,7 @@ def gsHttpToGs(path):
     return path.replace("https://storage.cloud.google.com/", "gs://").split("?")[0] # "?" is for possible "?authuser=0"
 
 
-# In[57]:
+# In[56]:
 
 
 def gsLs(path):
@@ -641,7 +665,7 @@ def gsLs(path):
     return ['gs://' + bucket + '/' + blob.name for blob in client.list_blobs(prefix=prefix)]
 
 
-# In[58]:
+# In[57]:
 
 
 def gsExists(path):
@@ -651,7 +675,7 @@ def gsExists(path):
     return client.blob(prefix).exists()
 
 
-# In[59]:
+# In[58]:
 
 
 def gsUploadFile(gsPath, path):
@@ -659,7 +683,7 @@ def gsUploadFile(gsPath, path):
     blob.upload_from_filename(path)
 
 
-# In[60]:
+# In[59]:
 
 
 def gsTxtread(path):
@@ -671,7 +695,7 @@ def gsTxtread(path):
     return blob.download_as_string().decode()
 
 
-# In[61]:
+# In[60]:
 
 
 def gsTxtwrite(path, s):
@@ -679,7 +703,7 @@ def gsTxtwrite(path, s):
     return blob.upload_from_string(s)
 
 
-# In[62]:
+# In[61]:
 
 
 def gsImwrite(path, img):
@@ -688,7 +712,7 @@ def gsImwrite(path, img):
     blob.upload_from_string(encimg.tobytes())
 
 
-# In[63]:
+# In[62]:
 
 
 def gsImread(path):
@@ -697,7 +721,7 @@ def gsImread(path):
     return cv2.imdecode(raw, 1)[:, :, [2, 1, 0]]
 
 
-# In[64]:
+# In[63]:
 
 
 def gsPklWrite(path, obj):
@@ -705,7 +729,7 @@ def gsPklWrite(path, obj):
     return blob.upload_from_string(pickle.dumps(obj, protocol=4))
 
 
-# In[65]:
+# In[64]:
 
 
 def gsPklRead(path):
@@ -715,7 +739,7 @@ def gsPklRead(path):
 
 # # Pickle
 
-# In[66]:
+# In[65]:
 
 
 def pklwrite(path, obj):
@@ -728,7 +752,7 @@ def pklwrite(path, obj):
             return pickle.dump(obj, f, protocol=4)
 
 
-# In[67]:
+# In[66]:
 
 
 def pklread(path):
@@ -739,7 +763,7 @@ def pklread(path):
             return pickle.load(f)
 
 
-# In[68]:
+# In[67]:
 
 
 if is_testing():
@@ -752,7 +776,7 @@ if is_testing():
 
 # # Images
 
-# In[69]:
+# In[68]:
 
 
 def _imreadSsh(path):
@@ -783,7 +807,21 @@ def imsread(paths):
     assert isinstance(paths, list) or isinstance(paths, tuple), type(paths)
     return [imread(path) for path in paths]
 
-def imwrite(path, img, jpeg_quality=95):
+def imwrite(path=None, img=None, jpeg_quality=95, print_assumptions=True):
+    if path is None:
+        path = "./imwrite_out.png"
+        if print_assumptions:
+            print(f"Write to {path}")
+    if type(img).__name__ == "Tensor":
+        if print_assumptions:
+            print("Convert to image")
+        if len(img.shape) != 4:
+            raise RuntimeError("Tensor needs 4 dimensions.")
+
+        img = (img - img.min()) / (img.max() - img.min())
+        img = (img * 255).numpy().transpose([0, 2, 3, 1]).astype(np.uint8)
+        img = lx.imGallery(img)
+
     if "gs://" in path:
         gsImwrite(path, img)
     os.makedirs(os.path.dirname(path), exist_ok=True)
@@ -819,7 +857,7 @@ def impad(img, top=0, bottom=0, left=0, right=0, color=255, mode='constant'):
         raise RuntimeError(f"Mode '{mode}' not found")
 
 
-# In[70]:
+# In[69]:
 
 
 def imscaleNN(img, s):
@@ -831,7 +869,7 @@ def imscaleBic(img, s):
 
 
 
-# In[71]:
+# In[70]:
 
 
 def imshow(array, scale=1, force_nb=False):
@@ -846,14 +884,14 @@ def imshow(array, scale=1, force_nb=False):
         cv2.waitKey()
 
 
-# In[72]:
+# In[71]:
 
 
 def imNewNoise(height, width, scaleNN=1):
     return imscaleNN((255 * np.random.rand(height, width, 3)).astype(np.uint8), scaleNN)
 
 
-# In[73]:
+# In[72]:
 
 
 if is_testing():
@@ -861,7 +899,7 @@ if is_testing():
     imshow(impad(img, bottom=10, right=10, mode='reflect'))
 
 
-# In[74]:
+# In[73]:
 
 
 def imCropUniformBorder(img, color=255):
@@ -900,7 +938,7 @@ if is_testing():
                         raise RuntimeError('Crop did not work for ', i, j, k, l)
 
 
-# In[75]:
+# In[74]:
 
 
 def imDecodeBytes(imBytes):
@@ -911,7 +949,7 @@ def imDecodeBytes(imBytes):
     return img
 
 
-# In[76]:
+# In[75]:
 
 
 #if is_testing():
@@ -919,14 +957,14 @@ def imDecodeBytes(imBytes):
 #    imshow(img)
 
 
-# In[77]:
+# In[76]:
 
 
 #if is_testing():
 #    imshow('http://via.placeholder.com/70.png')
 
 
-# In[78]:
+# In[77]:
 
 
 if is_testing():
@@ -945,7 +983,7 @@ if is_testing():
         imshow(out_large_interbic)
 
 
-# In[79]:
+# In[78]:
 
 
 def imRepeat(img, height, width):
@@ -959,7 +997,7 @@ def imRepeat(img, height, width):
     return img
 
 
-# In[80]:
+# In[79]:
 
 
 if is_testing():
@@ -967,33 +1005,33 @@ if is_testing():
     imshow(imRepeat(img_random_color_3x5, 7, 11), scale=10)
 
 
-# In[81]:
+# In[80]:
 
 
 def imNewWhite(height, width):
     return np.ones((height, width, 3)).astype(np.uint8) * 255
 
 
-# In[82]:
+# In[81]:
 
 
 if is_testing(): imshow(imNewWhite(10, 500) - 30)
 
 
-# In[83]:
+# In[82]:
 
 
 def imNewBlack(height, width):
     return np.zeros((height, width, 3)).astype(np.uint8)
 
 
-# In[84]:
+# In[83]:
 
 
 if is_testing(): imshow(imNewBlack(10, 500))
 
 
-# In[85]:
+# In[84]:
 
 
 def imAddNoiseGauss(img, std):
@@ -1005,7 +1043,7 @@ def imAddNoiseGauss(img, std):
 if is_testing(): imshow(imAddNoiseGauss(imNewWhite(150, 150), 30))
 
 
-# In[86]:
+# In[85]:
 
 
 if is_testing():
@@ -1013,7 +1051,7 @@ if is_testing():
     imshow('/tmp/img.jpg')
 
 
-# In[87]:
+# In[86]:
 
 
 def imJpgDegradation(img, quality):
@@ -1026,7 +1064,7 @@ def imJpgDegradation(img, quality):
 #if is_testing(): imshow(imJpgDegradation(imread('http://via.placeholder.com/150.png'), 15))
 
 
-# In[88]:
+# In[87]:
 
 
 def imCropCenter(img, size):
@@ -1041,35 +1079,35 @@ def imCropCenter(img, size):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[89]:
+# In[88]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 10))
 
 
-# In[90]:
+# In[89]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 49))
 
 
-# In[91]:
+# In[90]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 50))
 
 
-# In[92]:
+# In[91]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 100))
 
 
-# In[93]:
+# In[92]:
 
 
 def imCropRandom(img, size, auto_size=True):
@@ -1092,7 +1130,7 @@ def imCropRandom(img, size, auto_size=True):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[94]:
+# In[93]:
 
 
 #if is_testing(): 
@@ -1102,7 +1140,7 @@ def imCropRandom(img, size, auto_size=True):
 #    imshow(imCropRandom(imread('http://via.placeholder.com/40.png'), 41))
 
 
-# In[95]:
+# In[94]:
 
 
 def draw_rect(img, h_beg, h_end, w_beg, w_end, line_width, color=(0, 0, 0)):
@@ -1126,7 +1164,7 @@ def draw_rect(img, h_beg, h_end, w_beg, w_end, line_width, color=(0, 0, 0)):
 
 def imCropRandomSeed(img, sizes, idx, crop_idx, use_crop=True, use_minimap=False, mark_color=(200, 50, 50), scale=2, minimap_region=None):
     if img is None:
-        return imNewWhite(sizes[0], sizes[1])
+        return imNewWhite(sizes[0] * scale, sizes[1] * scale)
     
     def get_beg_end(beg_crop, size_crop, size_region, size_img):
         center = beg_crop + (size_crop // 2)
@@ -1184,7 +1222,7 @@ def imCropRandomSeed(img, sizes, idx, crop_idx, use_crop=True, use_minimap=False
         return crop
 
 
-# In[96]:
+# In[95]:
 
 
 if is_testing():
@@ -1192,13 +1230,19 @@ if is_testing():
     imshow(imCropRandomSeed(img, (10, 20), idx=0, crop_idx=0, use_crop=True, use_minimap=True, minimap_region=3), scale=3)
     imshow(imCropRandomSeed(img, (10, 20), idx=0, crop_idx=0, use_crop=False, use_minimap=True, minimap_region=3), scale=3)
     imshow(imCropRandomSeed(img, (10, 20), idx=0, crop_idx=0, use_crop=True, use_minimap=False), scale=3)
-    imshow(imCropRandomSeed(img, (10, 20), idx=0, crop_idx=0, use_crop=True, use_minimap=False), scale=3)
+    imshow(imCropRandomSeed(None, (10, 20), idx=0, crop_idx=0, use_crop=True, use_minimap=False), scale=3)
 
 
-# In[97]:
+# In[96]:
 
 
-def imGallery(imgs, pad=0, n_cols=None, n_rows=None):
+def imGallery(imgs, pad=0, n_cols=None, n_rows=None, ratio=None):
+    def get_nh_nw_from_ratio(n, height, width, ratio):
+        # (nw * widht) / (nh * height) = ratio
+        nw = int(np.clip(round(np.sqrt(n * height / width * ratio)), 1, n))
+        nh = int(np.clip(np.ceil(n / nw), 1, n))
+        return nh, nw
+    
     imgs = [imread(img) for img in imgs] # load if path or url
     
     n = len(imgs)
@@ -1212,6 +1256,9 @@ def imGallery(imgs, pad=0, n_cols=None, n_rows=None):
             raise RuntimeError("Can only fix rows or columns")
         nh = n_rows
         nw = int(np.ceil(n / nh))
+    elif ratio is not None:
+        height, width, _ = imgs[0].shape
+        nh, nw = get_nh_nw_from_ratio(n, height, width, ratio)
     else:
         nw = int(np.ceil(np.sqrt(n)))
         nh = int(np.ceil(n / nw))
@@ -1239,18 +1286,11 @@ def imGallery(imgs, pad=0, n_cols=None, n_rows=None):
     return out
 
 
-# In[98]:
+# In[97]:
 
 
 if is_testing():
-    # List of images from back to white
     listOfDummyImagesBlackToWhite = [imNewBlack(10, 10) + 25 * i for i in range(10)]
-
-
-# In[99]:
-
-
-if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:1]))
     imshow(imGallery(listOfDummyImagesBlackToWhite[:2]))
     imshow(imGallery(listOfDummyImagesBlackToWhite[:2], pad=1))
@@ -1259,9 +1299,18 @@ if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1))
     imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1, n_cols=3))
     imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1, n_rows=2))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1, n_rows=2))
+    
+    print("ratio")
+    listOfDummyImagesBlackToWhite = [imNewBlack(10, 3) + 5 * i for i in range(40)]
+    imshow(imGallery(listOfDummyImagesBlackToWhite, pad=1))
+    imshow(imGallery(listOfDummyImagesBlackToWhite, pad=1, ratio=1))
+    imshow(imGallery(listOfDummyImagesBlackToWhite, pad=1, ratio=16/9))
+    imshow(imGallery(listOfDummyImagesBlackToWhite, pad=1, ratio=10000))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:3], pad=1, ratio=0))
 
 
-# In[100]:
+# In[98]:
 
 
 #if is_testing():
@@ -1269,7 +1318,7 @@ if is_testing():
 #    imshow(imGallery([img40x20, img40x20, img40x20], pad=1))
 
 
-# In[101]:
+# In[99]:
 
 
 def imAddTextHeader(img, text, h=None, size=None):
@@ -1304,9 +1353,17 @@ if is_testing():
     imshow(imAddTextHeader(imNewBlack(100, 100), 'black image'))
 
 
+# In[100]:
+
+
+if is_testing():
+    imgs = [imNewWhite(5, 5) * 8 // i for i in range(9)]
+    imshow(imGallery(imgs))
+
+
 # # Video
 
-# In[102]:
+# In[101]:
 
 
 #def imgsToVid(imgs, file_name='out.avi'):
@@ -1326,20 +1383,23 @@ if is_testing():
 #imgsToVid(frames, 'test.avi')
 
 
-# In[103]:
+# In[102]:
 
 
 #ffmpeg -framerate 1/5 -i img%03d.png -c:v libx264 -r 30 -pix_fmt yuv420p out.mp4
 
+# Use padding for uneven
+# ffmpeg -framerate 1/5 -pattern_type glob -i '*.png' -c:v libx264 -r 30 -pix_fmt yuv420p -vf "crop=trunc(iw/2)*2:trunc(ih/2)*2"  ../out.mp4
 
-# In[104]:
+
+# In[103]:
 
 
 def imsToMp4(path, imgs, time_per_image, frame_rate_output_video=30, verbose=0):
     with tempfile.TemporaryDirectory() as d:
         for idx, img in enumerate(imgs):
             imwrite(os.path.join(d, f"img{idx:09d}.png"), img)
-        out = bash(f"cd {d} && ls && ffmpeg -framerate {1/time_per_image} -i img%09d.png -c:v libx264 -r 30 -pix_fmt yuv420p {path} -y")
+        out = bash(f"cd {d} && ls && ffmpeg -framerate {1/time_per_image} -i img%09d.png -c:v libx264 -r {frame_rate_output_video} -pix_fmt yuv420p {path} -y")
         if verbose:
             print(out)
 
@@ -1352,7 +1412,7 @@ if is_testing():
 
 # # pytorch
 
-# In[105]:
+# In[104]:
 
 
 def ptT(img_np_uint8):
@@ -1369,7 +1429,7 @@ def ptUint8(img_tensor):
     
 
 
-# In[106]:
+# In[105]:
 
 
 if is_testing():
@@ -1381,7 +1441,7 @@ if is_testing():
 
 # # Files
 
-# In[107]:
+# In[106]:
 
 
 def _fiFindByWildcardSsh(wildcard):
@@ -1402,7 +1462,7 @@ def fiFindImgs(dir_path, recursive=False, ext='png'):
     return fiFindByWildcard(wildcard)
 
 
-# In[108]:
+# In[107]:
 
 
 import os
@@ -1425,7 +1485,7 @@ if is_testing():
     print(pexpanduser('~/tmp/file.txt'))
 
 
-# In[109]:
+# In[108]:
 
 
 if is_testing():
@@ -1454,7 +1514,7 @@ if is_testing():
         imshow(imGallery(imgs))
 
 
-# In[110]:
+# In[109]:
 
 
 def fiRemove(path):
@@ -1482,7 +1542,7 @@ def fiAgeHours(path):
     return (time.time() - os.stat(path)[stat.ST_MTIME]) / 3600
 
 
-# In[111]:
+# In[110]:
 
 
 if is_testing():
@@ -1514,7 +1574,7 @@ if is_testing():
         print("files: " + "\n".join(fiFindByWildcard(os.path.join(d, "**", "*"))))
 
 
-# In[112]:
+# In[111]:
 
 
 def fiUnzip(zip_path, out_dir=None, remove=False):
@@ -1529,7 +1589,7 @@ def fiUnzip(zip_path, out_dir=None, remove=False):
         zip.extractall(out_dir)
 
 
-# In[113]:
+# In[112]:
 
 
 def fiJoin(*args):
@@ -1541,7 +1601,7 @@ if is_testing():
 
 # # Jupyter
 
-# In[114]:
+# In[113]:
 
 
 def jnSettings():
@@ -1552,7 +1612,7 @@ if is_testing():
     jnSettings()
 
 
-# In[115]:
+# In[114]:
 
 
 def jnNbFullWidth():
@@ -1566,7 +1626,7 @@ def jnPandasSettings():
     pd.set_option('display.max_colwidth', 100000)
 
 
-# In[116]:
+# In[115]:
 
 
 def jnImageRenderingPixelated():
@@ -1578,7 +1638,7 @@ def jnImageRenderingAuto():
     display(HTML('<style type = text/css> img { image-rendering: auto; } </style>'))
 
 
-# In[117]:
+# In[116]:
 
 
 def jnDefault():
@@ -1596,7 +1656,7 @@ if is_testing():
 
 
 
-# In[118]:
+# In[117]:
 
 
 def jnHtmlRow(strList, width=100):
@@ -1609,7 +1669,7 @@ if is_testing():
     
 
 
-# In[123]:
+# In[118]:
 
 
 def jnImgToHtmlB64(img):
@@ -1670,7 +1730,7 @@ def jnImageRow(imgs, names=None, width=100, external=False, title_img=None, titl
     display(HTML(html))
 
 
-# In[124]:
+# In[119]:
 
 
 if is_testing():
@@ -1679,7 +1739,7 @@ if is_testing():
    
 
 
-# In[125]:
+# In[120]:
 
 
 if is_testing():
@@ -1689,7 +1749,7 @@ if is_testing():
     jnImageRow(imgs, external=True, title='test')
 
 
-# In[126]:
+# In[121]:
 
 
 
@@ -1697,7 +1757,7 @@ if is_testing():
    jnImageRow(imgs=[imNewBlack(20, 20), imNewBlack(40, 20), imNewBlack(20, 800)])
 
 
-# In[127]:
+# In[122]:
 
 
 #def dfDisplayWithImgs(df):
@@ -1713,7 +1773,7 @@ if is_testing():
 
 # # GIT
 
-# In[128]:
+# In[123]:
 
 
 def gitGetHash():
@@ -1723,7 +1783,7 @@ if is_testing():
     print(gitGetHash())
 
 
-# In[129]:
+# In[124]:
 
 
 def gitGetBranch():
@@ -1733,7 +1793,7 @@ if is_testing():
     print(gitGetBranch())
 
 
-# In[130]:
+# In[125]:
 
 
 def gitGetMessage():
@@ -1743,7 +1803,7 @@ if is_testing():
     print(gitGetMessage())
 
 
-# In[131]:
+# In[126]:
 
 
 def gitGetAllFiles():
@@ -1753,7 +1813,7 @@ if is_testing():
     print(gitGetAllFiles())
 
 
-# In[132]:
+# In[127]:
 
 
 def gitListByTime(path=None):
@@ -1769,7 +1829,7 @@ if is_testing():
     print(gitListByTime()[-3:])
 
 
-# In[133]:
+# In[128]:
 
 
 def gitTime(path):
@@ -1780,7 +1840,7 @@ if is_testing():
     print(gitTime("__init__.py"))
 
 
-# In[134]:
+# In[129]:
 
 
 def gitTimeRaw(path):
@@ -1791,7 +1851,7 @@ if is_testing():
     print(gitTimeRaw("__init__.py"))
 
 
-# In[135]:
+# In[130]:
 
 
 def shellSortDirsBySize():
@@ -1803,14 +1863,14 @@ if is_testing():
 
 # # Package Files
 
-# In[136]:
+# In[131]:
 
 
 if is_testing():
     package_files = {}
 
 
-# In[137]:
+# In[132]:
 
 
 def shell(cmd):
@@ -1818,7 +1878,7 @@ def shell(cmd):
     return result.stdout.decode()
 
 
-# In[138]:
+# In[133]:
 
 
 if is_testing():
@@ -1850,7 +1910,7 @@ setup(
 )""".format(requirements=", ".join(["'" + r + "'" for r in requirements]), version=version)
 
 
-# In[139]:
+# In[134]:
 
 
 if is_testing():
@@ -1860,7 +1920,7 @@ description-file = README.md
 """
 
 
-# In[140]:
+# In[135]:
 
 
 if is_testing():
@@ -1883,14 +1943,14 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE."""
 
 
-# In[141]:
+# In[136]:
 
 
 if is_testing():
     package_files['README.md'] = ""
 
 
-# In[145]:
+# In[137]:
 
 
 if is_testing():
@@ -1901,7 +1961,7 @@ if is_testing():
     print(output.strip(), 'Version correct: ', output.strip() == version, "(", output.strip(), version, ")")
 
 
-# In[146]:
+# In[138]:
 
 
 if is_testing():
@@ -1912,7 +1972,7 @@ lx/__init__.py
 """
 
 
-# In[147]:
+# In[139]:
 
 
 if is_testing():
